#include "multitaskviewmanager.h"

#include <KF5/KGlobalAccel/KGlobalAccel>
#include <KF5/KWindowSystem/KWindowSystem>

#include <QAction>
#include <QApplication>
#include <QQmlContext>

#include "desktopbackground.h"
#include "icontheme.h"
#include "windowthumbnail.h"

namespace MultitaskView {

MultitaskViewManager::MultitaskViewManager()
{
    registerQmlTypes();

    registerShortcut();

    initDBusService();
}

MultitaskViewManager::~MultitaskViewManager()
{
    dbusService_.reset();

    close();
}

void MultitaskViewManager::grabbedKeyboardEvent(QKeyEvent *e)
{
    QApplication::sendEvent(view_->rootObject(), e);
}

void MultitaskViewManager::open()
{
    if (isOpen)
    {
        return;
    }

    initView();

    KWin::effects->grabKeyboard(this);

    isOpen = true;
}

void MultitaskViewManager::close()
{
    if (!isOpen)
    {
        return;
    }

    KWin::effects->ungrabKeyboard();

    destroyView();

    isOpen = false;
}

void MultitaskViewManager::initModel()
{
    model_.reset(new MultitaskViewModel());
}

void MultitaskViewManager::registerQmlTypes()
{
    qmlRegisterType<Plasma::WindowThumbnail>("org.kde.plasma", 2, 0, "WindowThumbnail");
    qmlRegisterType<DesktopBackground>("org.ukui.kwin", 2, 0, "DesktopBackground");
    qmlRegisterType<IconTheme>("IconTheme", 1, 0, "IconTheme");
}

void MultitaskViewManager::registerShortcut()
{
    auto* action = new QAction(this);
    action->setObjectName("ShowMultitaskView");
    action->setText("Show Multitask View");

    QKeySequence keySeq(Qt::META + Qt::Key_Tab);
    KGlobalAccel::self()->setDefaultShortcut(action, QList<QKeySequence>() << keySeq);
    KGlobalAccel::self()->setShortcut(action, QList<QKeySequence>() << keySeq);

    QObject::connect(action, &QAction::triggered, this, &MultitaskViewManager::open);
}

void MultitaskViewManager::initDBusService()
{
    dbusService_.reset(new DBusService(this));
}

void MultitaskViewManager::initView()
{
    view_.reset(new QQuickView());

    view_->setFlag(Qt::WindowType::BypassWindowManagerHint);

    KWindowSystem::setType(view_->winId(), NET::Dock);

    initModel();

    view_->rootContext()->setContextProperty("multitaskViewManager", this);
    view_->rootContext()->setContextProperty("multitaskViewModel", model_.get());

    view_->setSource(QUrl("qrc:/qml/multitaskview.qml"));

    // 必须在加载了QML之后才能设置view_对象的大小
    view_->setGeometry(KWin::effects->virtualScreenGeometry());

    connectSignals();

    view_->show();
}

void MultitaskViewManager::destroyView()
{
    if (view_ == nullptr)
    {
        return;
    }

    if (view_->isVisible())
    {
        view_->hide();
    }

    // 1. 由于可以在QML中通过JS调用到close函数，
    // 因此，不能直接销毁view_对象；必须延迟
    // 销毁该对象。
    // 2. model_对象可以立刻销毁。但是，为了对称
    // 和简化逻辑，需要保证view_对象和model_对象的
    // 互相依赖的关系。因此，同时延迟销毁model_对象。
    // 3. MultitaskViewManager的析构函数需要确保
    // view_对象和model_对象完成销毁。为了简化逻辑，
    // MultitaskViewManager的析构函数选择直接调用
    // destroyView函数，即采用了延迟销毁的做法。
    view_.release()->deleteLater();

    model_.release()->deleteLater();
}

void MultitaskViewManager::connectSignals()
{
    const auto* model = model_.get();
    const auto* rootObject = view_->rootObject();
    connect(model, SIGNAL(appWindowCreated(int)),
            rootObject, SIGNAL(qmlAppWindowCreated(int)));
    connect(model, SIGNAL(appWindowRemoved(int)),
            rootObject, SIGNAL(qmlAppWindowRemoved(int)));
    connect(model, SIGNAL(appWindowActivated()),
            rootObject, SIGNAL(qmlAppWindowActivated()));
    connect(model, SIGNAL(appWindowDesktopChanged(int, int, int)),
            rootObject, SIGNAL(qmlAppWindowDesktopChanged(int, int, int)));
    connect(model, SIGNAL(desktopRemoved(int)),
            rootObject, SIGNAL(qmlDesktopRemoved(int)));
    connect(model, SIGNAL(desktopAppended(int)),
            rootObject, SIGNAL(qmlDesktopAppended(int)));
    connect(model, SIGNAL(desktopMoved(int, int)),
            rootObject, SIGNAL(qmlDesktopMoved(int, int)));
    connect(model, SIGNAL(currentDesktopChanged(int, int)),
            rootObject, SIGNAL(qmlCurrentDesktopChanged(int, int)));
    connect(model, SIGNAL(screenCountChanged(int, int)),
            rootObject, SIGNAL(qmlScreenCountChanged(int, int)));
    connect(model, SIGNAL(screenSizeChanged()),
            rootObject, SIGNAL(qmlScreenSizeChanged()));
}

} // namespace MultitaskView
