#include "multitaskviewmodel.h"

#include <algorithm>

#include <QDesktopWidget>

namespace MultitaskView {

MultitaskViewModel::MultitaskViewModel(QObject *parent) : QObject(parent)
{
    updateScreenCount();

    updateModelData();

    connectSignals();
}

MultitaskViewModel::~MultitaskViewModel() = default;

QList<QVariant> MultitaskViewModel::getAppWindowList(int desktopIndex, int screenIndex) const
{
    if (!validateDesktopIndex(desktopIndex)
            ||!validateScreenIndex(screenIndex))
    {
        return QList<QVariant>();
    }

    auto desktopIter = std::find_if(desktopList_.cbegin(),
                                    desktopList_.cend(),
                                    [desktopIndex](const QVariant& desktop)
    {
        return desktop.value<Desktop>().index() == desktopIndex;
    });
    if (desktopIter == desktopList_.cend())
    {
        return QList<QVariant>();
    }

    const auto& desktop = desktopIter->value<Desktop>();
    const auto& screenList = desktop.screenList();
    auto screenIter = std::find_if(screenList.cbegin(),
                                   screenList.cend(),
                                   [screenIndex](const QVariant& screen)
    {
        return screen.value<Screen>().index() == screenIndex;
    });
    if (screenIter == screenList.cend())
    {
        return QList<QVariant>();
    }

    const auto& screen = screenIter->value<Screen>();
    auto appWindowList = screen.windowList();
    std::reverse(appWindowList.begin(), appWindowList.end());

    return appWindowList;
}

void MultitaskViewModel::removeAppWindow(int windowId)
{
    auto* window = KWin::effects->findWindow(windowId);
    if (window == nullptr)
    {
        return;
    }

    window->closeWindow();
}

void MultitaskViewModel::moveAppWindowToDesktop(int windowId,
                                                int srcDesktopInex,
                                                int destDesktopInex) const
{
    auto* window = KWin::effects->findWindow(windowId);
    if (window == nullptr
            || window->isOnAllDesktops()
            || !validateDesktopIndex(srcDesktopInex)
            || !validateDesktopIndex(destDesktopInex))
    {
        return;
    }

    auto desktopIndexList = window->desktops();
    if (!desktopIndexList.contains(srcDesktopInex))
    {
        return;
    }

    desktopIndexList.remove(desktopIndexList.indexOf(srcDesktopInex));

    if (desktopIndexList.contains(destDesktopInex))
    {
        return;
    }

    desktopIndexList.append(destDesktopInex);

    KWin::effects->windowToDesktops(window, desktopIndexList);
}

void MultitaskViewModel::activateAppWindow(int windowId)
{
    auto* window = KWin::effects->findWindow(windowId);
    if (window == nullptr)
    {
        return;
    }

    KWin::effects->activateWindow(window);
}

QIcon MultitaskViewModel::getWindowIcon(int windowId)
{
    auto id = static_cast<WId>(windowId);
    auto* window = KWin::effects->findWindow(id);
    if(window == nullptr)
    {
        return QIcon("");
    }

    return window->icon();
}

QList<QVariant> MultitaskViewModel::getDesktopList() const
{
    return desktopList_;
}

void MultitaskViewModel::removeDesktop(int desktopIndex)
{
    auto desktopCount = KWin::effects->numberOfDesktops();
    if (desktopCount <= 1)
    {
        return;
    }

    if (!validateDesktopIndex(desktopIndex))
    {
        return;
    }

    const auto oldDesktopIndex = desktopIndex;
    if (desktopIndex == 1)
    {
        moveDesktopWindows(1, 2);

        desktopIndex = 2;
    }

    for (; desktopIndex <= desktopCount; desktopIndex++)
    {
        moveDesktopWindows(desktopIndex, desktopIndex - 1);
    }

    KWin::effects->setNumberOfDesktops(desktopCount - 1);

    updateModelData();

    emit desktopRemoved(oldDesktopIndex);
}

void MultitaskViewModel::appendDesktop()
{
    auto desktopCount = KWin::effects->numberOfDesktops();
    if (desktopCount >= MAX_DESKTOP_COUNT)
    {
        return;
    }

    KWin::effects->setNumberOfDesktops(desktopCount + 1);

    updateModelData();

    emit desktopAppended(desktopCount + 1);
}

void MultitaskViewModel::moveDesktop(int srcDesktopIndex, int destDesktopIndex)
{
    if (!validateDesktopIndex(srcDesktopIndex)
            || !validateDesktopIndex(destDesktopIndex))
    {
        return;
    }

    if (srcDesktopIndex == destDesktopIndex)
    {
        return;
    }

    const auto desktopCount = KWin::effects->numberOfDesktops();
    KWin::effects->setNumberOfDesktops(desktopCount + 1);

    auto tempDesktopIndex = desktopCount + 1;
    moveDesktopWindows(destDesktopIndex, tempDesktopIndex);
    // 将临时虚拟桌面的数据暂时更新到模型数据中，
    // 方便后面调用moveDesktopWindows函数将
    // 临时虚拟桌面中的窗口移动到来源虚拟桌面
    updateModelData();

    moveDesktopsOneStep(createDesktopIndexList(srcDesktopIndex, destDesktopIndex));

    moveDesktopWindows(tempDesktopIndex, srcDesktopIndex);

    KWin::effects->setNumberOfDesktops(desktopCount);

    updateModelData();

    emit desktopMoved(srcDesktopIndex, destDesktopIndex);
}

int MultitaskViewModel::getCurrentDesktop()
{
    return KWin::effects->currentDesktop();
}

void MultitaskViewModel::setCurrentDesktop(int desktopIndex)
{
    if (!validateDesktopIndex(desktopIndex))
    {
        return;
    }

    KWin::effects->setCurrentDesktop(desktopIndex);
}

QRect MultitaskViewModel::getFullScreenGeometry(int desktopIndex, int screenIndex) const
{
    return KWin::effects->clientArea(KWin::clientAreaOption::FullScreenArea,
                                     screenIndex, desktopIndex);
}

bool MultitaskViewModel::hasModalChildWindow(int windowId)
{
    auto* window = KWin::effects->findWindow(windowId);
    if (window == nullptr)
    {
        return false;
    }

    return window->findModal() != nullptr;
}

void MultitaskViewModel::onWindowAdded(KWin::EffectWindow *window)
{
    if(!shouldAcceptWindow(window))
    {
        if(!window->isModal()) {
            return;
        }
    }

    updateModelData();

    emit appWindowCreated(getWindowId(window));
}

void MultitaskViewModel::onWindowDeleted(KWin::EffectWindow *window)
{
    updateModelData();

    emit appWindowRemoved(getWindowId(window));
}

void MultitaskViewModel::onWindowActivated(KWin::EffectWindow* window)
{
    // KWin在发送windowActivated信号时
    // window对象是有可能为空的。对于多任务
    // 视图来说，只关心真正有窗口被激活的清空。
    // 因此，如果window对象为空，则不转发
    // 该信号
    if (window == nullptr)
    {
        return;
    }

    emit appWindowActivated();
}

void MultitaskViewModel::onDesktopPresenceChanged(KWin::EffectWindow *window,
                                                  int oldDesktopIndex, int newDesktopIndex)
{
    if (!shouldAcceptWindow(window))
    {
        return;
    }

    updateModelData();

    emit appWindowDesktopChanged(getWindowId(window), oldDesktopIndex, newDesktopIndex);
}

void MultitaskViewModel::onNumberScreensChanged()
{
    updateModelData();

    const auto oldScreenCount = screenCount_;

    updateScreenCount();

    emit screenCountChanged(oldScreenCount, screenCount_);
}

void MultitaskViewModel::updateScreenCount()
{
    screenCount_ = KWin::effects->numScreens();
}

// static
bool MultitaskViewModel::isExtensionMode()
{
    QDesktopWidget desktop;
    for (int index = 0; index < desktop.numScreens(); index++)
    {
        if(KWin::effects->virtualScreenGeometry()
                == desktop.screenGeometry(index))
        {
            return false;
        }
    }

    return true;
}

int MultitaskViewModel::logicalScreenCount() const
{
    const auto screenCount = KWin::effects->numScreens();
    if (screenCount == 1
            || !isExtensionMode())
    {
        return 1;
    }

    return screenCount;
}

QList<QVariant> MultitaskViewModel::createWindowList(int desktopIndex, int screenIndex)
{
    if (!validateDesktopIndex(desktopIndex)
            || !validateScreenIndex(screenIndex))
    {
        return QList<QVariant>();
    }

    QList<QVariant> windowList;

    const auto& effectWindowList = KWin::effects->stackingOrder();
    for (auto iter = effectWindowList.cbegin();
         iter != effectWindowList.cend(); iter++)
    {
        auto* effectWindow = *iter;
        if (!shouldAcceptWindow(effectWindow))
        {
            continue;
        }

        if (!effectWindow->desktops().contains(desktopIndex)
                || effectWindow->screen() != screenIndex)
        {
            continue;
        }

        Window window(getWindowId(effectWindow),
                      effectWindow->caption(),
                      effectWindow->pos().x(),
                      effectWindow->pos().y(),
                      effectWindow->decorationInnerRect().width(),
                      effectWindow->decorationInnerRect().height(),
                      effectWindow->isMinimized());

        windowList.append(QVariant::fromValue(window));
    }

    return windowList;
}

void MultitaskViewModel::updateModelData()
{
    QList<QVariant> desktopList;

    const auto desktopCount = KWin::effects->numberOfDesktops();
    for (int desktopIndex = 1; desktopIndex <= desktopCount; desktopIndex++)
    {
        QList<QVariant> screenList;
        for (int screenIndex = 0; screenIndex < logicalScreenCount(); screenIndex++)
        {
            auto screen = QVariant::fromValue(Screen(screenIndex, createWindowList(desktopIndex, screenIndex)));
            screenList.append(std::move(screen));
        }

        auto desktop = QVariant::fromValue(Desktop(desktopIndex, screenList));
        desktopList.append(std::move(desktop));
    }

    desktopList_ = std::move(desktopList);
}

void MultitaskViewModel::moveDesktopWindows(int srcDesktopIndex, int destDesktopIndex)
{
    if (!validateDesktopIndex(srcDesktopIndex)
            || !validateDesktopIndex(destDesktopIndex))
    {
        return;
    }

    auto srcDesktopIter = std::find_if(desktopList_.cbegin(),
                                       desktopList_.cend(),
                                       [srcDesktopIndex](const QVariant& desktop)
    {
       return desktop.value<Desktop>().index() == srcDesktopIndex;
    });
    if (srcDesktopIter == desktopList_.cend())
    {
        return;
    }

    const auto& desktop = srcDesktopIter->value<Desktop>();
    const auto& screenList = desktop.screenList();
    for (auto screenIter = screenList.cbegin();
         screenIter != screenList.cend(); screenIter++)
    {
        const auto& screen = screenIter->value<Screen>();
        const auto& windowList = screen.windowList();
        for (auto windowIter = windowList.begin();
             windowIter != windowList.end(); windowIter++)
        {
            const auto& window = windowIter->value<Window>();
            moveAppWindowToDesktop(window.id(), srcDesktopIndex, destDesktopIndex);
        }
    }
}

void MultitaskViewModel::moveDesktopsOneStep(const QVector<int>& desktopIndexList)
{
    if (desktopIndexList.size() < 2)
    {
        return;
    }

    for (auto iter = desktopIndexList.cbegin();
         iter != desktopIndexList.cend() -1; iter++)
    {
        const auto destDesktopIndex = *iter;
        const auto srcDesktopIndex = *(iter + 1);
        moveDesktopWindows(srcDesktopIndex, destDesktopIndex);
    }
}

// static
WId MultitaskViewModel::getWindowId(KWin::EffectWindow *window)
{
    return window->parent()->property("windowId").toULongLong();
}

// static
QVector<int> MultitaskViewModel::createDesktopIndexList(int srcDesktopIndex, int destDesktopIndex)
{
    QVector<int> desktopIndexList;

    if (!validateDesktopIndex(srcDesktopIndex)
            || !validateDesktopIndex(destDesktopIndex))
    {
        return desktopIndexList;
    }

    if (srcDesktopIndex > destDesktopIndex)
    {
        for (auto index = destDesktopIndex; index <= srcDesktopIndex; index++)
        {
            desktopIndexList.append(index);
        }
    }
    else
    {
        for (auto index = destDesktopIndex; index >= srcDesktopIndex; index--)
        {
            desktopIndexList.append(index);
        }
    }

    return desktopIndexList;
}

// static
bool MultitaskViewModel::validateDesktopIndex(int desktopIndex)
{
    return desktopIndex >= 1 && desktopIndex <= MAX_DESKTOP_COUNT;
}

bool MultitaskViewModel::validateScreenIndex(int screenIndex) const
{
    return screenIndex >= 0 && screenIndex < logicalScreenCount();
}

// static
bool MultitaskViewModel::shouldAcceptWindow(KWin::EffectWindow* window)
{
    return !window->isDeleted()
            && window->isNormalWindow()
            && !window->isSkipSwitcher()
            && window->isOnCurrentActivity()
            && window->acceptsFocus();
}

void MultitaskViewModel::connectSignals()
{
    QObject::connect(KWin::effects, &KWin::EffectsHandler::windowAdded,
                     this, &MultitaskViewModel::onWindowAdded);
    QObject::connect(KWin::effects, &KWin::EffectsHandler::windowDeleted,
                     this, &MultitaskViewModel::onWindowDeleted);
    QObject::connect(KWin::effects, &KWin::EffectsHandler::windowActivated,
                     this, &MultitaskViewModel::onWindowActivated);
    QObject::connect(KWin::effects, &KWin::EffectsHandler::desktopPresenceChanged,
                     this, &MultitaskViewModel::onDesktopPresenceChanged);
    QObject::connect(KWin::effects, SIGNAL(desktopChanged(int, int)),
                     this, SIGNAL(currentDesktopChanged(int, int)));
    QObject::connect(KWin::effects, &KWin::EffectsHandler::numberScreensChanged,
                     this, &MultitaskViewModel::onNumberScreensChanged);
    QObject::connect(KWin::effects, &KWin::EffectsHandler::virtualScreenSizeChanged,
                     this, &MultitaskViewModel::screenSizeChanged);
}

} // namespace MultitaskView
